﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Advent.TrsRmt;            // for TRSSIM remote control wrapper
using System.Reflection;

namespace TrsRmt_CSharp_Example
{
    public partial class MainForm : Form
    {

        /// <summary>
        /// This is our remote control object which we use to connect to TRSSIM
        /// </summary>
        private TrsRmt Remote = new TrsRmt();


        public MainForm()
        {
            InitializeComponent();
            // Add an event handler for when this form closes
            this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.MyFormClosing);

            // We add our event selectors directly to the combo box
            cbo_EventSelector.Items.Clear();
            foreach (Enum E in Enum.GetValues(typeof(EventSelector)))
            {
                cbo_EventSelector.Items.Add(E);
            }
            cbo_EventSelector.SelectedIndex = 0;    


            // Setup Tab stops

            int[] Q = new int [20];
            int i=0;
            for(i=0;i<20; i++) Q[i] = 30 + 80 * i;

            rtb_EventInfo.SelectionTabs = Q;
            rtb_DeviceInfo.SelectionTabs = Q;
            rtb_Script.SelectionTabs = Q;
        }

        private void MyFormClosing(object sender, EventArgs e)
        {
            // ensure we are disconnected before this program ends
            Disconnect();
        }


        /// <summary>
        /// This attempts to connect to the TRSSIM server
        /// </summary>
        private void TryConnect()
        {
            try
            {
                // If we are already connected to nothing
                if (Remote.IsConnected()) return;

                // We atttempt to connect to the TRSSIM server
                Remote.Connect(txt_ServerIP.Text, int.Parse(txt_Port.Text));

                // If we made it this far we must be connected
                btn_Connect.Enabled = false;
                MainTab.Enabled = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
           }
        }

        private void Disconnect()
        {
            Remote.Disconnect();
            btn_Connect.Enabled = true;
            MainTab.Enabled = false;
        }

        private void btn_Connect_Click(object sender, EventArgs e)
        {
            TryConnect();
        }


        /// <summary>
        /// This extracts logs from TRSSIM and puts each of them in a tab control with a richtextbox that contains the log contents
        /// </summary>
        private void btn_GetLog_Click(object sender, EventArgs e)
        {
            try
            {

                List<Log> Logs = Remote.GetLogs(txt_LogFilter.Text);

                //Now we re-load each log into each tab 
                LogTab.TabPages.Clear();

                foreach (Log log in Logs)
                {
                    // Create a new tabpage and add it with the name of the log
                    TabPage tab = new TabPage(log.Name);
                    LogTab.TabPages.Add(tab);

                    // We create a new richtext box and add the contents of the log to it
                    RichTextBox R = new RichTextBox();
                    R.Text = log.Contents;
                    tab.Controls.Add(R);
                    //Resize the richtextbox so it fits well in the tab
                    R.Top = 0;
                    R.Left = 0;
                    R.Width = tab.ClientRectangle.Width;
                    R.Height = tab.ClientRectangle.Height;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        private void btn_Disconnect_Click(object sender, EventArgs e)
        {
            Disconnect();
        }

        /// <summary>
        /// This fetches all of the device information present in TRSSIM
        /// </summary>
        private void btn_GetDevices_Click(object sender, EventArgs e)
        {
            try
            {
                // Fetch all the device information from TRRSIM
                List<DeviceInfo> Devices = Remote.GetDevices();
                rtb_DeviceInfo.Text = "";

                // we use a stringbuilder to be more efficient
                StringBuilder Info = new StringBuilder();

                // Now we display each device in the richtext box
                foreach (DeviceInfo dev in Devices)
                {
                    
                    Info.AppendLine("DeviceID:\t" + dev.DeviceID);
                    Info.AppendLine("Identifier:\t" + dev.Identifier);
                    Info.AppendLine("Name:\t" + dev.Name);
                    Info.AppendLine("Serial Number:\t" + dev.SerialNumber);
                    Info.AppendLine("Software Version:\t" + dev.SoftwareVersion);
                    Info.AppendLine("Hardware Version:\t" + dev.HardwareVersion);
                    Info.AppendLine();
                }
                rtb_DeviceInfo.Text = Info.ToString();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        /// <summary>
        /// This instructs TRSSIM to open a file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btn_OpenFile_Click(object sender, EventArgs e)
        {
            try
            {
                Remote.OpenFile(txt_OpenPath.Text);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        /// <summary>
        /// This instructs TRSSIM to perform a reset
        /// </summary>
        private void btn_Reset_Click(object sender, EventArgs e)
        {
            try
            {
                Remote.Reset();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        /// <summary>
        /// This instructs TRRSIM to shutdown.
        /// </summary>
        private void btn_ShutDown_Click(object sender, EventArgs e)
        {
            try
            {
                // Confirm that we REALLY want to shutdown
                if (MessageBox.Show("You are about to shutdown TRSSIM. This will disconnect you and prevent you from reconnecting until TRSSIM is restarted. Continue?", "Confirm TRSSIM Shutdown", MessageBoxButtons.YesNo) != DialogResult.Yes) return;

                Remote.Exit();
                Disconnect();        // NO point in continuing
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        /// <summary>
        /// This sets the contents of the script editor to the text in the editor in this program
        /// </summary>
        private void btn_SetScriptSource_Click(object sender, EventArgs e)
        {
            try
            {
                //TODO: figure out how to change the RichTextBox control to use \r\n line ends
                Remote.SetScriptSource(rtb_Script.Text.Replace("\n", "\r\n"));
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }


        private void btn_DoStatements_Click(object sender, EventArgs e)
        {
            try
            {
                //TODO: figure out how to change the RichTextBox control to use \r\n line ends
                Remote.DoStatements(rtb_Script.Text.Replace("\n", "\r\n"));
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }


        private void btn_StartScript_Click(object sender, EventArgs e)
        {
            try
            {
                Remote.SendScriptCommand(ScriptCommand.RUN);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        private void btn_StopScript_Click(object sender, EventArgs e)
        {
            try
            {
                Remote.SendScriptCommand(ScriptCommand.END);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }
        }

        /// <summary>
        /// This simply updates the status display periodically
        /// </summary>
        private void timer_Update_Tick(object sender, EventArgs e)
        {

            timer_Update.Enabled = false;
            try
            {
                ScriptState ScriptStatus = ScriptState._NOSTATUS;
                ScriptState DoStatus = ScriptState._NOSTATUS;

                if (!Remote.IsConnected()) return;

                Remote.GetStatus(ref ScriptStatus, ref DoStatus);


                lbl_ScriptStatus.Text = "Status: Script=" + Enum.GetName(ScriptStatus.GetType(), ScriptStatus) + " Do=" + Enum.GetName(DoStatus.GetType(), DoStatus);
                timer_Update.Enabled = true;

            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }

        }

        private int GetDeviceID_For_EventCapableDevice()
        {
            // First we scan the available devices for an AI-5120 which can deal with events
            foreach (DeviceInfo d in Remote.GetDevices())
            {
                if (d.Name.Contains("AI-5120"))
                {
                    return  d.DeviceID;
                }
            }
            return -1;
        }



        /// <summary>
        /// This fetches an event from the first device which is capable of dealing with events ("AI-5120")
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btn_GetEvent_Click(object sender, EventArgs e)
        {
            try
            {

                rtb_EventInfo.Text = "";

                // First we scan the available devices for an AI-5120 which can deal with events
                int DeviceID = GetDeviceID_For_EventCapableDevice();
                if (DeviceID <0) throw new Exception("No devices with event capability are present in TRSSIM");

                // We ask the remote control to return all the event information
                Event Evnt = Remote.GetEvent(DeviceID,(EventSelector) cbo_EventSelector.SelectedItem);
                if (Evnt == null) throw new Exception("No such event exists");
                StringBuilder sb = new StringBuilder();

                sb.AppendLine("Event Name:\t" + Evnt.Name);
                sb.AppendLine("Event Type:\t" + Evnt.EventType);
                sb.AppendLine("Event Time:\t" + Evnt.Time);
                sb.AppendLine("PC Time:\t" + Evnt.PCTime);
                sb.AppendLine("ACWaveRefNum:\t" + Evnt.ACWaveRefNum);
                sb.AppendLine("DCWaveRefNum:\t" + Evnt.DCWaveRefNum);
                //Now we optionally print out all the field information
                if (Evnt.Fields.Count > 0)
                {
                    sb.AppendLine(Evnt.Fields.Count + " Fields");
                    foreach (EventField f in Evnt.Fields)
                    {
                        sb.AppendLine("\t" + f.Name + ":\t" + f.Value );
                    }
                }
                rtb_EventInfo.Text = sb.ToString();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }           
        }

        /// <summary>
        /// This fetches basic waveform information and displays the results in a richtext box
        /// </summary>
        private void btn_GetWaveformInfo_Click(object sender, EventArgs e)
        {
            try
            {
                // First we scan the available devices for an AI-5120 which can deal with events
                int DeviceID = GetDeviceID_For_EventCapableDevice();
                if (DeviceID < 0) throw new Exception("No devices with waveform capability are present in TRSSIM");

                WaveformInfo winfo = new WaveformInfo();
                // request the waveform info from the remote control dll
                Remote.GetWaveformInfo(DeviceID, int.Parse(txt_WaveformRefNum.Text), ref winfo);

                DisplayWaveformInfo(winfo);

            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }           
        }

        private void DisplayWaveformInfo(WaveformInfo w)
        {
            rtb_WaveformInfo.Text = "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine("DeviceID:\t" + w.DeviceID);
            sb.AppendLine("Waveform RefNum:\t" + w.WaveformRefNum);
            sb.AppendLine("NumSamples:\t" + w.NumSamples);
            sb.AppendLine("SampleRate:\t" + w.SampleRate);
            sb.AppendLine("StartTime:\t" + w.StartTime);
            sb.AppendLine("BitsPerSample:\t" + w.BitsPerSample);

            rtb_WaveformInfo.Text = sb.ToString();
        }

        private void btn_GetWaveform_Click(object sender, EventArgs e)
        {

            try
            {
                // First we scan the available devices for an AI-5120 which can deal with events
                int DeviceID = GetDeviceID_For_EventCapableDevice();
                if (DeviceID < 0) throw new Exception("No devices with waveform capability are present in TRSSIM");

                Waveform W = new Waveform();
                // request the waveform info from the remote control dll
                Remote.GetWaveform(DeviceID, int.Parse(txt_WaveformRefNum.Text), ref W);

                // display this waveform information
                DisplayWaveformInfo(W.Info);

                // now we must write the information to a file


                WriteWaveformToFile(W);

            }
            catch (Exception ex)
            {
                MessageBox.Show("Error:" + ex.Message);
            }

        }

        private void WriteWaveformToFile(Waveform W)
        {
            SaveFileDialog diag = new SaveFileDialog();
            diag.Title = "Save waveform to file";

            if (diag.ShowDialog() != DialogResult.OK) return;

            // Create a new text file with which to output the text
            System.IO.StreamWriter f = System.IO.File.CreateText(diag.FileName);

            try
            {
                f.WriteLine("Waveform");
                f.WriteLine("DeviceID:\t" + W.Info.DeviceID);
                f.WriteLine("Waveform RefNum:\t" + W.Info.WaveformRefNum);
                f.WriteLine("NumSamples:\t" + W.Info.NumSamples);
                f.WriteLine("SampleRate:\t" + W.Info.SampleRate);
                f.WriteLine("StartTime:\t" + W.Info.StartTime);
                f.WriteLine("BitsPerSample:\t" + W.Info.BitsPerSample);
                f.WriteLine("Samples");

                int i = 0; int N = 0;
                N = W.Samples.Length;
                for (i = 0; i < N; i++)
                {
                    f.WriteLine(W.Samples[i]);
                }
                f.Close();

            }
            catch (Exception ex)
            {
                f.Close();
                throw ex;
            }
        }








    }
}
