﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Runtime.InteropServices;       // For DLLImport
using System.Diagnostics;                   // for Stopwatch
using System.Net.Sockets;
using System.IO;
using Advent.TrsRmt;


namespace Advent.TrsRmt
{


    #region ("Enumerations")

    /// <summary>
    /// This enumerates the possible TRSSIM script states
    /// </summary>
    public enum ScriptState
    {
        _NOSTATUS = -1,		/* not a real value -> just means parameter not initialized*/
        STOP = 0,		/* no script program executing */
        RUN = 1,		    /* script is running */
        PAUSE = 2,		/* script is paused */
        IDLE = 3,		/* script is waiting for events */
        ERROR = 4,		/* error detected in script program */
    }

    /// <summary>
    /// This enumerates the TRSSIM script commands
    /// </summary>
    public enum ScriptCommand
    {
        END = 0,
        PAUSE = 1,
        RESUME = 2,
        RUN = 3,
    }

    /// <summary>
    /// This enumerates the event selecttors for GET EVENT
    /// </summary>
    public enum EventSelector
    {
        FIRST = 0,
        LAST = 1,
        NEXT = 2,
        PREV = 3,
        CURRENT = 4,
    }
    #endregion



    /// <summary>
    /// This class wraps the TRSSIM remote control DLL into an easier to use .Net object
    /// </summary>
    public class TrsRmt
    {

        /// <summary>
        /// This represents an exception caused by the trsrmt DLL returning an error code.
        /// </summary>
        public class TrsRmtException : Exception
        {
            public TrsRmtException(String Msg):base(Msg)
            {
            }
        }

        private const char STX = (char)2;  // start delimiter for all TRSSIM Remote commands
        private const char ETX = (char) 3;  // end delimiter for all TRSSIM remote commands

        private TcpClient _Tcp = null;          // TCP client connection
        private StreamWriter _Writer = null;   // Network stream for communication
        private StreamReader _Reader = null;   // Network stream for communication
        private  int _Timeout  = 20000;         // Timeout for communications in milliseconds


        /// <summary>
        /// This attempts to connect to the TRSSIM server. If it is unsuccessful it will throw an exception
        /// </summary>
        /// <param name="ServerIP">Dotted decimal IP address (ie 192.168.1.102)</param>
        /// <param name="Port">Port number used to Connect to TRSSIM (note: TRSSIM must be configured to accept connections on this port)</param>
        /// <returns></returns>
        public void Connect(String ServerIP, int Port)
        {
            Disconnect();
            // Connect to TRSSIM and extract streams which we can use for communications
            _Tcp = new TcpClient(ServerIP, Port);

            // Set the default timeout
            NetworkStream St = _Tcp.GetStream();
            St.WriteTimeout = _Timeout;
            St.ReadTimeout = _Timeout;
            _Writer = new StreamWriter(St);
            _Reader = new StreamReader(St);
        }

        /// <summary>
        /// This disconnects from the TRSSIM server
        /// </summary>
        public void Disconnect()
        {
            // Tear everything down
            if (_Writer != null) _Writer.Close();
            _Writer = null;
            if (_Reader != null) _Reader.Close();
            _Reader = null;
            if (_Tcp != null)_Tcp.Close();
            _Tcp = null;
        }

#region("Parsing and low level code")


        /// <summary>
        /// This function sends a particular command to TRSSIM with no parsing or error checking.
        /// </summary>
        private void Write(String Command)
        {
            if (_Writer == null) throw new TrsRmtException("Cannot send command " + Command + ": Not Connected");

            // We send the command to TRSSIM enclosed in STX and ETX characters
            _Writer.Write((char)STX + Command + (char)ETX);
            _Writer.Flush();
        }

        /// <summary>
        /// This sends a command to TRSSIM and performs some basic parsing on the response.
        /// This parses all responses enclosed within the STX/ETX characters. It does not search for WARNINGS or ERRORs
        /// </summary>
        /// <param name="Command">TRSSIM Command</param>
        /// <returns>Response as a string</returns>
        /// <exception cref="TrsRmtException">Thrown if error detected</exception>
        private String Send(String Command)
        {
            Write(Command); // Write the command to TRSSIM

            StringBuilder RX = new StringBuilder();
            char[] Buffer = new char[1000];
            Stopwatch Timer = new Stopwatch();
            Timer.Start();
            // Now we wait for a response enclosed in STX and ETX
            do
            {
                // Read data from the socket
                int Length = _Reader.Read(Buffer, 0, Buffer.Length);
                if (Length > 0)
                {
                    // We append the newly read data to the string
                    RX.Append(Buffer, 0, Length);
                }
         
                // Now search for the response delimiters
                String s = RX.ToString();
                int StartAt, StopAt;
                StartAt = s.IndexOf(STX);
                StopAt = s.IndexOf(ETX);
                if ((StartAt >=0) && (StopAt >=0))
                {
                    if (StopAt > StartAt)
                    {   
                        return s.Substring(StartAt + 1, StopAt - StartAt-1);
                    } else {
                        throw new TrsRmtException("Unexpected ETX STX sequence in response to command " + Command);
                    }
                }
            } while (Timer.ElapsedMilliseconds < _Timeout);
            throw new TimeoutException("Timed out while sending command " + Command);           
        }

        /// <summary>
        /// This sends a command to TRSSIM and checks the response for ERRORs or WARNINGS
        /// </summary>
        /// <param name="Command">TRSSIM Command</param>
        /// <returns>Response as an array of string lines</returns>
        /// <exception cref="TrsRmtException">Thrown if error detected</exception>
        string[] SendCommand(String Command)
        {
            const string ErrorIndicator = "ERROR";
            const string WarningIndicator = "WARNING";

            // We send the command and fetch the response
            String RX = Send(Command);
            
            // Now we check the response for pre-canned errors
            if (RX.StartsWith(ErrorIndicator)) 
                throw new TrsRmtException(RX.Substring((ErrorIndicator.Length+1)));
            if (RX.StartsWith(WarningIndicator)) 
                throw new TrsRmtException("WARNING: " + RX.Substring((WarningIndicator.Length+1)));
            return RX.Split(new String[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
        }


#endregion

        /// <summary>
        /// This sets the timeout for communications with the TRSSIM Server in seconds
        /// </summary>
        public void SetTimeout(int TimeoutSeconds)
        {
            _Timeout = TimeoutSeconds * 1000;
            if (_Tcp == null) return;
            NetworkStream St = _Tcp.GetStream();
            St.WriteTimeout = _Timeout;
            St.ReadTimeout = _Timeout;
        }

        /// <summary>
        /// This sets/gets the communications timeout value in milliseconds
        /// </summary>
        public int Timeout
        {
            get
            {
                return _Timeout;
            }

            set
            {
                _Timeout = value ;
                if (_Tcp == null) return;
                NetworkStream St = _Tcp.GetStream();
                St.WriteTimeout = _Timeout;
                St.ReadTimeout = _Timeout;

            }
        }

        /// <summary>
        /// This enumerates all the devices available on the TRSSIM remote and returns a list of the results.
        /// </summary>
        /// <returns></returns>
        public List<DeviceInfo> GetDevices()
        {
            List<DeviceInfo> L = new List<DeviceInfo>();

            // We query "GET DEVICES" from TRSSIM and fetch the response
            string[] Devs = SendCommand("GET DEVICES");
            if (Devs.Length < 1) throw new TrsRmtException("Too few responses to GET DEVICES");

            // We expect 'OK GET COUNT='n'
            // We parse each line after the first which should be "Devn= <info string>"
            int i = 0;
            for (i = 1; i < Devs.Length; i++)
            {
                L.Add(new DeviceInfo(Devs[i]));
            }
            return L;
        }

 


        /// <summary>
        /// This returns one or more logs available in TRSSIM. if LogName is null then this will return all logs available in TRSSIM. If LogName 
        /// is a non-null value then this will return the the log with the matching name
        /// </summary>
        /// <param name="SpecificLogName">Name of the specified log to return, or empty string for all logs</param>
        public List<Log> GetLogs(String LogName)
        {
            List<Log> L = new List<Log>();
            // We query TRSSIM for the specific log
            String[] RX = SendCommand("GET LOG " + LogName);
            /* Each log in the response should be 
                Log<n>=<Name>
                >First Line
                >Second Line */
            int i = 0;
            String Name=null;
            StringBuilder Contents = new StringBuilder();
            for (i = 1; i < RX.Length; i++)
            {
                if (RX[i].StartsWith("LOG"))
                {
                    // If we already have a log present then add it to the list of logs
                    if ((Name != null) && (Contents != null))
                    {
                        L.Add(new Log(Name, Contents.ToString()));
                        Name = "";
                        Contents = new StringBuilder();
                    }
                    int idx = RX[i].IndexOf('=');
                    if (idx > 0)
                    {
                        Name = RX[i].Substring(idx + 1);
                        Contents = new StringBuilder();
                    }
                }
                else if (RX[i].StartsWith(">"))
                {
                    // if the line starts with '>' then it is a line in the log and we append it as such
                    Contents.Append(RX[i].Substring(1) + "\r\n");
                }
            }
            // Add the last entry to the list of logs
            if ((Name != null) && (Contents != null))
            {
                L.Add(new Log(Name, Contents.ToString()));
                Name = "";
                Contents = new StringBuilder();
            }
            return L;
        }

        /// <summary>
        /// This returns True if we are connected to TRSSIM
        /// </summary>
        public bool IsConnected()
        {
            if (_Tcp == null) return false;
            return _Tcp.Connected;
        }


        /// <summary>
        /// This causes TRSSIM to perform a reset.
        /// </summary>
        public void Reset()
        {
            SendCommand("RESET");
        }

        /// <summary>
        /// This will cause TRSSIM to shutdown. Careful because you will not be able to re-establish a connection after this call unless you re-start TRSSIM manually.
        /// </summary>
        /// <remarks>This will terminate the connection to TRSSIM</remarks>
        public void Exit()
        {
            // We send the EXIT command without waiting for a response and quickly disconnect.
            Write("EXIT");
            Disconnect();
        }

        /// <summary>
        /// This executes the following script statements in TRSSIM. 
        /// </summary>
        /// <param name="ScriptStatements">Script statements to execute</param>
        /// <remarks>Note: this is completely separate from the Script portion of TRSSIM. See the DO command documentation for formatting.</remarks>
        public void DoStatements(String ScriptStatements)
        {
            SendCommand("DO " + ScriptStatements);
        }

        /// <summary>
        /// This opens the specified file in TRSSIM
        /// </summary>
        /// <param name="FileName">Filename to open</param>
        public void OpenFile(String FileName)
        {
            SendCommand("FILE OPEN " + FileName);
        }

        /// <summary>
        /// This sets the contents of the script editor in TRSSIM
        /// </summary>
        public void SetScriptSource(String Script)
        {
            SendCommand("SCRIPT SOURCE " + Script);
        }

        /// <summary>
        /// This controls the execution of the script by sending the specified script control command
        /// </summary>
        /// <param name="Command">Script command to execute</param>
        public void SendScriptCommand(ScriptCommand Command)
        {
            SendCommand("SCRIPT " + Enum.GetName(typeof(ScriptCommand),Command));
        }



        /// <summary>
        /// This returns an event object containing all available event information.
        /// </summary>
        /// <param name="DeviceId">ID corresponding to the device to read this event from</param>
        /// <param name="Selector">Selects which event to return</param>
        /// <returns>Event object if exists, null if no such event exists</returns>
        public Event GetEvent(int DeviceId, EventSelector Selector)
        {

            int EventType = 0;
            String Name = "";
            double TimeFromStart = 0;
            DateTime PCTime = DateTime.MinValue;
            int ACWaveRefNum = 0;
            int DCWaveRefNum = 0;
            List<EventField> F = new List<EventField>();        // event fiels

            // We Query TRSSIM for the specified event
            String[] RX = SendCommand("DEV " + (char) DeviceId + " GET EVENT " + Enum.GetName(typeof(EventSelector),Selector) );

            if (RX.Length <= 1) return null;

            /* Now we parse the response 
             * 
             * */
            /// Now we scan the response for the neccesary info
            int i = 0;
            for (i = 1; i < RX.Length; i++)
            {
                // Locate the '='  in this line
                int idx = RX[i].IndexOf('=');
                if (idx > 0)
                {
                    if (RX[i].StartsWith(">"))
                    {
                        //If it starts with '>' then it is an event field measurement
                        F.Add(new EventField(RX[i].Substring(1, idx-1),RX[i].Substring( idx+1)));
                    }
                    else
                    {
                        // Now we interpret the line based on name where NAME=<Value>
                        
                        switch (RX[i].Substring(0, idx))
                        {
                            case "TYPE":
                                String[] Values = RX[i].Substring(idx + 1).Split(':');
                                if (Values.Length < 2) throw new TrsRmtException("Too few fields in Event TYPE response " + RX[i]);
                                EventType = int.Parse(Values[0]);
                                Name = Values[1];
                                break;
                            case "TIME":
                                // Time is formatted  <yr> : <mn> : <day> : <hr> : <min> : <sec> : <time from start>
                                String[] T = RX[i].Substring(idx + 1).Split(':');
                                if (T.Length < 7) throw new TrsRmtException("Too few fields in Event TIME response " + RX[i]);

                                int Year = int.Parse(T[0]);
                                int Month = int.Parse(T[1]);
                                int Day = int.Parse(T[2]);
                                int Hour = int.Parse(T[3]);
                                int Minute = int.Parse(T[4]);
                                int Second = int.Parse(T[5]);
                                TimeFromStart = double.Parse(T[6]);
                                PCTime = new DateTime(Year, Month, Day, Hour, Minute, Second);
                                break;
                            case "WAVE":
                                // Wave is formatted WAVE= <AC coupled waveform ref #> : <DC coupled waveform ref #>
                                String[] W = RX[i].Substring(idx + 1).Split(':');
                                if (W.Length < 2) throw new TrsRmtException("Too few fields in Event WAVE response " + RX[i]);
                                ACWaveRefNum = int.Parse(W[0]);
                                DCWaveRefNum = int.Parse(W[1]);
                                break;
                        }
                    }
                }

            }
            return new Event(Name, EventType, TimeFromStart, PCTime, ACWaveRefNum, DCWaveRefNum, F);
        }


 

        /// <summary>
        /// This gets the current status of the TRSSIM script and last DO statement
        /// </summary>
        /// <param name="ScriptStatus">Current script execution status</param>
        /// <param name="DoStatus">Current status of the last DO statement</param>
        public void GetStatus(ref ScriptState ScriptStatus, ref ScriptState DoStatus)
        {
            ScriptStatus = ScriptState._NOSTATUS;
            DoStatus = ScriptState._NOSTATUS;

            String[] RX = SendCommand("GET STATUS");
            if (RX.Length <= 0) throw new TrsRmtException("Too few fields in GET STATUS response");

            // We split the string into chunks separated by '=' and <space>
            String[] X = RX[0].Split(new String[] { "=", " " }, StringSplitOptions.RemoveEmptyEntries);

            int ScriptIdx = RX[0].IndexOf("SCRIPT=");
            int DoIdx = RX[0].IndexOf("DO=");
            if ((ScriptIdx < 0) || (DoIdx < 0)) throw new TrsRmtException("Invalid response to GET STATUS: " + RX[0]);

            int i = 0;
            for (i = 0; i < X.Length; i++)
            {
                switch (X[i])
                {
                    case "SCRIPT":
                        if (i < (X.Length-1)) ScriptStatus = NameToScriptState(X[i+1]);
                        break;
                    case "DO":
                        if (i < (X.Length - 1)) DoStatus = NameToScriptState(X[i + 1]);
                        break;
                }
            }
        }

        /// <summary>
        /// This converts the name of a script state to an enum value
        /// </summary>
        /// <param name="ScriptStateName"></param>
        /// <returns></returns>
        private ScriptState NameToScriptState(String ScriptStateName)
        {
            foreach (ScriptState E in Enum.GetValues(typeof(ScriptState)))
            {
                if (String.Compare(Enum.GetName(typeof(ScriptState),E),ScriptStateName,true)==0) return E;
            }
            return ScriptState._NOSTATUS;
        }


        /// <summary>
        /// This returns basic waveform informatino for the specified device
        /// </summary>
        /// <param name="DeviceID">ID of the device associated with this waveform</param>
        /// <param name="WaveformRefNum">Reference number of the waveform to retrieve information on</param>
        /// <param name="Info">Structure filled with waveform information</param>
        public void GetWaveformInfo(int DeviceID,
                                    int WaveformRefNum,
                                    ref WaveformInfo Info)
        {
            Info.NumBlocks = 0;
            Info.BitsPerSample = 0;
            Info.DeviceID = DeviceID;
            Info.MaxBlockSize = 0;
            Info.NumSamples = 0;
            Info.SampleRate = 0;
            Info.ScaleFact = 0;
            Info.StartTime = 0;
            Info.WaveformRefNum = WaveformRefNum;

            // Query the waveform information from TRSSIM
            String[] RX = SendCommand("DEV " + (char)DeviceID + " GET WAVEINFO " + WaveformRefNum);
            foreach( String X in RX)
            {
                int idx = X.IndexOf('=');
                switch(X.Substring(0,idx).Trim())
                {
                    case "INFO":
                        // INFO = <start time> : <# samples> : <sample rate> : <scale factor> : <maximum block size> : <bits per sample>
                        String[] Q = X.Substring(idx+1).Split(':');
                        if (Q.Length < 6) throw new TrsRmtException("Too few fields in INFO response:" + X);

                        Info.StartTime = double.Parse(Q[0]);
                        Info.NumSamples = int.Parse(Q[1]);
                        Info.SampleRate = double.Parse(Q[2]);
                        Info.ScaleFact = double.Parse(Q[3]);
                        Info.MaxBlockSize = int.Parse(Q[4]);
                        Info.BitsPerSample = int.Parse(Q[5]);
                        break;
                    case "BLOCKS":
                        //BLOCKS = <number of data blocks required to the read the waveform>
                        Info.NumBlocks = int.Parse(X.Substring(idx+1));
                        break;
                }
            }
        }

        /// <summary>
        /// This download an entire waveform from TRSSIM.
        /// </summary>
        /// <param name="DeviceID">ID of the device to download from.</param>
        /// <param name="WaveformRefnum">Waveform ID of the waveform to download</param>
        /// <param name="Wave">Waveform object which will be filled with new waveform information</param>
        public void GetWaveform(int DeviceID, int WaveformRefnum, ref Waveform Wave)
        {
            /// First we fetch the actual waveform information
            GetWaveformInfo(DeviceID, WaveformRefnum, ref Wave.Info);
            if (Wave.Info.NumSamples <= 0)
            {
                Wave.Samples = new double[0];
                return;
            }

            // Make enoungh storage space for this waveform
            Wave.Samples = new double[Wave.Info.NumSamples];
            int SampleIndex = 0;


            for (int i = 1; i <= Wave.Info.NumBlocks; i++)
            {
                if (SampleIndex > Wave.Info.NumSamples) break;
                // Issue the command to extract the specified waveform block
                String[] RX = SendCommand("DEV " + (char) DeviceID + " GET WAVEBLOCK " + WaveformRefnum + " " + i);

                /* We expect 
                    OK .....
                    BLOCK = <#sample returned> : <#bits per sample> : <checksum>
                    > (first line of 32 samples in hexadecimal format)
                    > (second line of 32 samples in hexadecimal format)
                    ...
                    > (last line of up to 32 samples in hexadecimal format>
                */

                const String BlockDef = "BLOCK=";
                if (!RX[1].StartsWith(BlockDef)) throw new TrsRmtException("Expected GET WAVEBLOCK to contain " + BlockDef + " but got " + RX[1]);
                String[] BlockFields = RX[1].Substring(BlockDef.Length).Split(':');
                if (BlockFields.Length < 2) throw new TrsRmtException("Expected " + BlockDef + " to have at least 2 fiels but got " + BlockFields.Length);
                int BitsPerSample = int.Parse(BlockFields[1]);
                if (BitsPerSample != 16) throw new TrsRmtException("Cannot download unsupported resolution of " + BitsPerSample + " bits per sample");

                int ExpectedSamples = int.Parse(BlockFields[0]);


                if (RX.Length < 2) throw new TrsRmtException("Too few responses to GET WAVEBLOCK");

                // TODO: check the resolution of the samples being returned here
                // Convert each line of hex characters

                for (int j = 2; j < RX.Length; j++)
                {
                    if (SampleIndex > Wave.Info.NumSamples) break;
                    if (!RX[j].StartsWith(">")) throw new TrsRmtException("Expected WAVEBLOCK response line " + j + " to start with '>'");

                    String Line = RX[j].Substring(1).Trim();
                    
                    for (int k = 0; k < Line.Length; k += 4)
                    {
                        if (SampleIndex > Wave.Info.NumSamples) break;

                        // Convert this hex number to a 16-bit integer and then to a double using the scale factor specified
                        Wave.Samples[SampleIndex] = Wave.Info.ScaleFact * (double) Int16.Parse(Line.Substring(k, 4), System.Globalization.NumberStyles.HexNumber);
                        SampleIndex++;
                        ExpectedSamples--;
                    }
                }

                if (ExpectedSamples != 0)
                {
                    if (ExpectedSamples > 0) throw new TrsRmtException("Block " + i + " of waveform data contained " + ExpectedSamples + " too few samples");
                    if (ExpectedSamples < 0) throw new TrsRmtException("Block " + i + " of waveform data contained " + ExpectedSamples + " too many samples");
                }

            }
        }
    }
}